from .attrdict.dictionary import AttrDict
from .insee import _insee_idb_data_url, _insee_varcode_data_url, _insee_data, find_closest_idbank
from .processing import get_dict_insee_mods, write_compress_json, read_decompress_json
from .dataframes import insee_modalites_as_dataframe, filter_modalites, modalitesHelper



from pandas import DataFrame
from .attrdict.dictionary import AttrDict
from zlib import compress, decompress
from json import dumps, loads

def parse_insee_mods(idb : DataFrame) -> AttrDict:
    insee_mods = {}

    for idx, row in idb.iterrows():
        _mods = row.list_mod.split('.')
        _root = insee_mods
        for m in _mods:
            if m in _root:
                    _root = _root[m]
            else:
                    _root[m] = {}
                    _root = _root[m]


    return AttrDict(insee_mods)


def add_idbanks(idb: DataFrame, insee_mods : AttrDict) -> AttrDict:
    for idx, row in idb.iterrows():
        _mods = row.list_mod.split('.')
        _m = insee_mods
        for m in _mods[:-1]:
            _m = _m.get(m)

        if not len(_m.get(_mods[-1])):
            _m[_mods[-1]]["_data"] = [row.idbank]
        elif _m.get(_mods[-1]) == {}:
            _m[_mods[-1]]["_data"] = [row.idbank]
        elif isinstance(_m[_mods[-1]], dict) and len(_m[_mods[-1]].keys()):
            _m[_mods[-1]]["_data"] = [row.idbank]
        else:
            _m[_mods[-1]]["_data"].append(row.idbank)
    return insee_mods

def get_dict_insee_mods(idb: DataFrame) -> AttrDict:
    mods = parse_insee_mods(idb)
    mods = add_idbanks(idb, mods)
    return mods

def write_compress_json(attrdict : AttrDict, path: str) -> None:
    cjson = compress(dumps(attrdict).encode('utf-8'))
    if path.suffix == '.cjson':
        _path = path
    else:
        _path = path
        _path.suffix = '.cjson'
    with open(path, 'wb') as fp:
        fp.write(cjson)


def read_decompress_json(path: str) -> None:
    with open(path, 'rb') as fp:
        djson = loads(decompress(fp.read()).decode('utf-8'))
    return AttrDict(djson)


from importlib.resources import files
from pathlib import Path
from typing import Union
from functools import cache
from lxml import etree
import pandas as pd
import requests as rq
import difflib as dfl


DATA_PATH = files("pybdm_insee").parent.joinpath("data/")\
    if isinstance(files("pybdm_insee"), Path)\
    else files("pybdm_insee")._paths.pop().parent.joinpath("data/")

def _insee_varcode_data_url(insee_row):
    
    return f"""
    https://bdm.insee.fr/series/sdmx/data/{insee_row.VAR_CODE}/.......{insee_row.MOD_CODE}.
    """

def _insee_idb_data_url(idbank: str):
    return f"""
    https://api.insee.fr/series/BDM/V1/data/SERIES_BDM/{idbank}
    """.strip()


@cache
def _insee_data(translate=True):
    res = {
        "variables" : pd.read_csv(DATA_PATH.joinpath("202211_liste_variables_modalites.csv"), sep =";", compression={'method': 'gzip'}),
        "naf" : pd.read_csv(DATA_PATH.joinpath("naf2008-listes-completes-5-niveaux.csv"), sep =";", compression={'method': 'gzip'}),
        "idbank": pd.read_csv(DATA_PATH.joinpath("202211_correspondance_idbank_dimension.csv"), sep =";", compression={'method': 'gzip'})
    }
    return res

def idb_exists(_idbank : str):
    _fidbank = _idbank.zfill(9)
    idb = _insee_data().get("idbank")
    return idb[idb.idbank.apply(lambda v : str(v).zfill(9)) == _fidbank].any().any()

def find_closest_idbank(idbank : str, n=5):
    idb = _insee_data().get("idbank")
    dfl.get_close_matches(idbank, idb.idbank.unique().tolist(), n=n)

def insee_get_access_token():
    r = rq.post(
        "https://api.insee.fr/token",
        data="grant_type=client_credentials",
        headers={"Authorization" : "Basic a2tEb3BGWEh0d1hXUGN1Tzh5NWJ0QlZGQWw4YTowaWZOVTA3SEtfU3VDV3lhMnlJSWQ1WFQ2TEFh"},
        )

    return r.json()['access_token']


def insee_bdm_get(idbank: Union[str, int]):
    """Tries to mimic the following cURL command:

    curl -X GET
    --header 'Accept: application/xml'
    --header 'Authorization: Bearer ce39f356-d3f3-3b99-838e-8326267d8488'
    'https://api.insee.fr/series/BDM/V1/data/SERIES_BDM/001656506'

    Args:
        idbank (Union): the idbank either str or int, will be padded to 9 chars
    """
    idbank = str(idbank).zfill(9)
    token = insee_get_access_token()

    r = rq.get(
        url=_insee_idb_data_url(idbank),
        headers={
            'Accept': 'application/xml',
            'Authorization': f'Bearer {token}'
        }
    )

    return r.text

def process_xml_output(_xmlstr : str):
    doc = etree.XML(_xmlstr.encode('utf-8'))
    ss = doc.find(".//Series")
    res = [val.items() for val in ss.iter()]
    obj = {
        "metadata":{ k:v for k,v in res[0] },
        "series":[
            {k:v for k,v in row} for row in res[1:]
        ]
    }

    return obj


import pandas as pd
import json

from pybdm_insee.tools.insee import (
    process_xml_output, _insee_idb_data_url,
    insee_bdm_get, process_xml_output, _insee_data,
    DATA_PATH as dpath
)

with open(dpath.joinpath('insee_modalites_2022.json'), 'r') as fp : mods = json.load(fp)

data = _insee_data()
idb = data["idbank"]

modcols = idb.list_mod.map(lambda v: v.split('.'))

ncols = max([len(row) for row in modcols])

dfmod = pd.DataFrame.from_records(data=modcols, columns=[f"var{i}" for i in range(ncols)])

dfmod.idbank = idb.idbank
dfmod.family = idb.famille


import platform
from subprocess import call
from pkgutil import iter_modules

cmdline = {
    "Linux":"curl -sSL https://install.python-poetry.org | python3 -",
    "Darwin":"curl -sSL https://install.python-poetry.org | python3 -",
    "Windows":"(Invoke-WebRequest -Uri https://install.python-poetry.org -UseBasicParsing) | python3 -"
}

def run_poetry_install():
        # call(cmdline[platform.system()], shell=True)
        pass


import click
import json
import sys
import pandas as pd
from subprocess import call
from importlib.util import find_spec

from pybdm_insee.tools.insee import (
    process_xml_output,
    insee_bdm_get, process_xml_output,
    _insee_data, idb_exists, find_closest_idbank
)

from pybdm_insee.installs.poetry_install import run_poetry_install


obj = process_xml_output(insee_bdm_get("001656506"))
df = pd.DataFrame.from_records(obj["series"])

def echo_json_pager(json_series):
    def _generate_output():
        for row in json_series:
            yield row
    click.echo_via_pager(_generate_output())

@click.group()
def cli():
    """
    Hi ! This is a small CLI and python toolbox to help people query INSEE databases easily.
    Two commands as of the latest release:

    - ask : opens a prompt and helps the user figure out which series they are interested in through iterative querying\n
    - fetch : uses either the IDBANK identifiers, or the so-called "modality" values to identify the series, fetches them and saves or display them.
    """
    pass

@cli.command()
def ask():
    click.echo('Nothing :/')

@cli.command()
@click.argument('pkg')
def install(pkg):
    click.secho(f'Attempting to install package {pkg}...', fg="blue")
    if pkg == 'spacy':
        try:
            # cmdline = ["python", "-m", "spacy", "download", "en_core_web_md"]
            # call(cmdline, shell=True, executable=sys.executable)
            from spacy.cli.download import download
            download("en_core_web_md")
        except Exception as e:
            click.secho(f"Please install Poetry at https://python-poetry.org/docs/", fg="yellow", bold=True, err=True)
            raise e

@cli.command()
@click.option('--idbank', default=0,
    help="""
    A unique series ID used by INSEE to identify data.
    See https://api.insee.fr/catalogue/site/themes/wso2/subthemes/insee/pages/item-info.jag?name=BDM&version=V1&provider=insee
    for more information. Please use other options to query using different means. If you wish to explore the datasets,
    please use the `ask` feature
    """, type=click.STRING)
@click.option('--mod', default="",
    help="""
    Each series belongs to a hierarchy of more or less 9 categories and subcategories called "modalités" (modality).
    If you know the modality's name, it restricts the number of series you might care about.
    There are two approaches which can be further specified using the `--ask` option : either you want additional information
    to narrow down series to a few idbanks, or you want a bulk query which gives you multiple tables. 
    """, type=click.STRING)
@click.option('--out', default="cli",
    help="""
    Whether you want the output to go to the console or to be stored in a file,
    Any value other than 'cli' will be taken as the path+name of the file, defaulting to csv storage.
    You can still set the 'clifmt' if your output is cli, it will simply change the way the result is displayed
    """, type=click.STRING)
@click.option('--clifmt', default="df", help='What format to print in the terminal', type=click.STRING)
@click.option('--format', default="csv", help='Whether to store the result as csv, json or pickle', type=click.STRING)
@click.option('--sep', default=";", help='If output is csv, separator to use.', type=click.STRING)
@click.option('--ask', is_flag=True, default=False, help='Whether to run the `ask` command if you input wrong information.')
def fetch(idbank, mod, out, clifmt, format, sep, ask):
    """
    This script will fetch the specified series using the INSEE BDM service using 2022 codes.
    Returns data either as csv, json or pickle for storage, or as dataframe or json for display.
    For more information: https://api.insee.fr/catalogue/site/themes/wso2/subthemes/insee/pages/item-info.jag?name=BDM&version=V1&provider=insee)
    """
    if idbank != '0':
        if idb_exists(idbank):
            obj = process_xml_output(insee_bdm_get(idbank))
            df = pd.DataFrame.from_records(obj["series"])

            if clifmt == 'json':
                echo_json_pager(obj["series"])
            elif clifmt == 'df':
                click.echo(df)
            else:
                click.secho('WARNING: cli format incorrect !', fg='red', blink=True, bold=True, err=True)
                click.secho('Printing df as default..', fg='green', bold=True, err=True)
                click.echo(df)
        else:
            click.secho('WARNING: IDBANK value incorrect !', fg='red', blink=True, bold=True, err=True)
            
            if not ask:
                click.secho('NB: use either the `ask` command or the --ask flag if you want suggestions :)', fg='green', bold=True, err=True)
            else:
                matches = find_closest_idbank(idbank)
                click.secho('NB: Here are the closest 5 idbanks, use `ask` to figure what they are ;)', fg='green', bold=True, err=True)
                click.echo(matches)
    else:
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit()

cli.add_command(fetch)
cli.add_command(ask)

if __name__ == "__main__":
    cli()


from .tools.insee import (
    process_xml_output, _insee_idb_data_url,
    insee_get_access_token, insee_bdm_get,
    process_xml_output, find_closest_idbank
)
from .tools.attrdict.dictionary import AttrDict
from .tools.processing import get_dict_insee_mods, write_compress_json, read_decompress_json
from .tools.dataframes import insee_modalites_as_dataframe, filter_modalites, modalitesHelper
from .installs.poetry_install import run_poetry_install


import pandas as pd
import json
from typing import Optional, Union, List, Dict

from pybdm_insee.tools.insee import _insee_data, DATA_PATH as dpath


def insee_modalites_as_dataframe():
    data = _insee_data()
    idb = data["idbank"]

    modcols = idb.list_mod.map(lambda v: v.split('.'))

    ncols = max([len(row) for row in modcols])

    dfmod = pd.DataFrame.from_records(data=modcols, columns=[f"var{i}" for i in range(ncols)])

    dfmod["idbank"] = idb.idbank
    dfmod["family"] = idb.famille

    return dfmod

def filter_modalites(
    data : Union[str, List[str], Dict[int,str]] = [],
    no_var : Optional[Union[int, List[int]]] = None,
    var_value : Optional[Union[str, List[str]]] = None,
    block_missing : Optional[bool] = False,
    reference : pd.DataFrame = None
    ) -> pd.DataFrame:
    """
    Takes either a single column value, a list of column values, or a dictionary of column values.
    There is an option to pick the col numbers are the col values separately using two arguments (no_var, var_value)

    Args:
        data (Union[str, List, Dict]): Either the value of var0, the values of var0, var1, ..., varN (N=len(List)),
            or a dict of (index, value) pairs
        no_var (Optional[int]): a single value as a number or a list of number,
            this field is used (as an option) to specify on which columns we filter,
            cannot be used without var_value
        var_value (Optional[str]): a list of the values each specified column should take
        block_missing (Optional[bool], optional): if the function should raise in case of wrong input. Defaults to False.

    Returns:
       pd.DataFrame: the sliced dataframe filtered on the given columns & values
    """
    df = insee_modalites_as_dataframe() if reference is None else reference

    if not len(data):
        if no_var is None or var_value is None:
            if block_missing:
                raise NotImplementedError("Only two options to give data, none were used.")
            else:
                return df
        else:
            no_var = no_var if isinstance(no_var, list) else [no_var]
            var_value = var_value if isinstance(var_value, list) else [var_value]
            return df[
                eval(
                    " & ".join([
                        f"(df['var{i}'] == '{val}')" for i, val in zip(no_var, var_value)
                        ])
                    )
                ]
    else:
        if isinstance(data, str):
            return df[df.var0 == data]
        elif isinstance(data, list):
            return df[
                eval(
                    " & ".join([
                        f"(df['var{i}'] == '{val}')" for i, val in enumerate(data)
                        ])
                    )
                ]
        elif isinstance(data, dict):
            return df[
                eval(
                    " & ".join([
                        f"(df['var{i}'] == '{val}')" for i, val in data.items()
                        ])
                    )
                ]

class modalitesHelper:
    """A helper class to guide people to the right INSEE series
    """
    def __init__(
        self,
        filter_data : Union[str, List[str], Dict[int,str]] = [],
        filter_no_var : Optional[Union[int, List[int]]] = None,
        filter_var_value : Optional[Union[str, List[str]]] = None,
        block_missing : Optional[bool] = False
        ) -> pd.DataFrame:
        """
        A helper which outputs an english language explanation of what a column means and the corresponding values.
        It is possible to ask the helper 
        Takes either a single column value, a list of column values, or a dictionary of column values.
        There is an option to pick the col numbers are the col values separately using two arguments (no_var, var_value)

        Args:
            data (Union[str, List, Dict]): Either the value of var0, the values of var0, var1, ..., varN (N=len(List)),
                or a dict of (index, value) pairs
            no_var (Optional[int]): a single value as a number or a list of number,
                this field is used (as an option) to specify on which columns we filter,
                cannot be used without var_value
            var_value (Optional[str]): a list of the values each specified column should take
            block_missing (Optional[bool], optional): if the function should raise in case of wrong input. Defaults to False.

        Returns:
        pd.DataFrame: the sliced dataframe filtered on the given columns & values
        """
        self.df = filter_modalites(data=filter_data, no_var=filter_no_var, var_value=filter_var_value)
        self.descriptions = pd.read_csv(dpath.joinpath("202211_liste_variables_modalites.csv.gz"), sep =";", compression={'method': 'gzip'})

    def filter(
        self,
        data : Union[str, List[str], Dict[int,str]] = [],
        no_var : Optional[Union[int, List[int]]] = None,
        var_value : Optional[Union[str, List[str]]] = None,
        block_missing : Optional[bool] = False
        ) -> pd.DataFrame:
        """
        Allows for recursive filtering of the dataframe to help people narrow down the options

        Args:
            data (Union[str, List[str], Dict[int,str]], optional): Similar to `filter_modalites`. Defaults to [].
            no_var (Optional[Union[int, List[int]]], optional): Similar to `filter_modalites`. Defaults to None.
            var_value (Optional[Union[str, List[str]]], optional): Similar to `filter_modalites`. Defaults to None.
            block_missing (Optional[bool], optional): Similar to `filter_modalites`. Defaults to False.

        Returns:
            pd.DataFrame: returns self to allow chaining (hlp.filter(..).filter(..) etc)
        """
        self.df = filter_modalites(data=data, no_var=no_var, var_value=var_value, reference=self.df)

        return self

    def reset(self):
        self.df = insee_modalites_as_dataframe()

    def to_dataframe(self):
        return self.df

    def ask_column_values(self, numbers : Union[int, List[int]]):
        if isinstance(numbers, int):
            col = self.df[f"var{numbers}"]
            values = col.unique().tolist()
            return values
        elif isinstance(numbers, list):
            return {
                nb: self.df[f"var{nb}"].values.uniques().to_list() for nb in numbers
            }
    
    def ask_column_desc_by_name(self, name : str):
        desc = self.descriptions[self.descriptions.MOD_CODE == name].MOB_LIBEN
        return desc if desc.size > 1 else desc.item()

    def ask_column_desc_by_number(self, nb : int):
        names = self.ask_column_values(nb)
        desc = self.descriptions.loc[self.descriptions.MOD_CODE.isin(names), ["MOD_CODE", "MOB_LIBEN"]]
        return desc

    def ask_desc(self, _input : Union[str, int]):
        return self.ask_column_desc_by_name(_input) if isinstance(_input, str) else self.ask_column_desc_by_number(_input)

    def ask(self, query : Union[str, int, Dict[Union[str, int], str]]):
        """_summary_

        Args:
            query (Union[str, int, Dict[Union[str, int], str]]): A dict with either names or numbers as keys, and as a value
                whether one wants help on the values the column takes or the description. Accepts a single number or name aswell.
        """
        if isinstance(query, str):
            return self.ask_column_desc_by_name(query)
        elif isinstance(query, int):
            return self.ask_column_values(query)
        elif isinstance(query, dict):
            is_asking_names = (lambda k,v :  False if isinstance(k,str) else v not in ['d','desc', 'description'])
            return { k : self.ask_column_values(k) if is_asking_names(k,v) else self.ask(k) for k,v in query.items() }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybdm_insee',
 'pybdm_insee.installs',
 'pybdm_insee.tools',
 'pybdm_insee.tools.attrdict']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'colorama>=0.4.6,<0.5.0',
 'lxml>=4.9.1,<5.0.0',
 'pandas>=1.5.1,<2.0.0',
 'requests>=2.28.1,<3.0.0']

extras_require = \
{'caching': ['spacy>=3.4.3,<4.0.0'], 'spacy': ['spacy>=3.4.3,<4.0.0']}

entry_points = \
{'console_scripts': ['insee = pybdm_insee.main:cli']}

setup_kwargs = {
    'name': 'pybdm-insee',
    'version': '0.1.6',
    'description': '',
    'long_description': '# My very first package !\n\nThis is a very simple package helping people use the [IBDM api from french INSEE](https://api.insee.fr/catalogue/site/themes/wso2/subthemes/insee/pages/item-info.jag?name=BDM&version=V1&provider=insee).',
    'author': "Arno's Stuff",
    'author_email': 'bcda0276@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
